from abc import abstractmethod
from enum import Enum
from functools import partial, wraps
from typing import (
    Callable,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from ranzen import some
import torch
from torch import Tensor

__all__ = [
    "Comparator",
    "accuracy",
    "accuracy_per_class",
    "accuracy_per_group",
    "accuracy_per_subclass",
    "balanced_accuracy",
    "fscore",
    "fscore_per_class",
    "fscore_per_group",
    "fscore_per_subclass",
    "groupwise_metric",
    "hard_prediction",
    "macro_fscore",
    "max_difference_1d",
    "merge_indices",
    "nanmax",
    "nanmin",
    "pdist_1d",
    "precision_at_k",
    "robust_accuracy",
    "robust_fscore",
    "robust_fscore_gap",
    "robust_gap",
    "robust_ovr_tpr",
    "robust_tnr",
    "robust_tpr",
    "subclass_balanced_accuracy",
    "subclasswise_metric",
    "tnr_per_subclass",
    "tpr_differences",
    "tpr_per_subclass",
    "weighted_nanmean",
]


@torch.no_grad()  # pyright: ignore
def hard_prediction(logits: Tensor) -> Tensor:
    logits = logits.squeeze(1) if logits.ndim == 2 else torch.atleast_1d(logits)
    return (logits > 0).long() if logits.ndim == 1 else logits.argmax(dim=1)


@torch.no_grad()  # pyright: ignore
def precision_at_k(
    y_pred: Tensor, *, y_true: Tensor, top_k: Union[int, Tuple[int, ...]] = (1,)
) -> List[Tensor]:
    """Computes the accuracy over the k top predictions for the specified values of k"""
    if isinstance(top_k, int):
        top_k = (top_k,)
    maxk = max(top_k)
    batch_size = y_true.size(0)
    _, pred = y_pred.topk(k=maxk, dim=1, largest=True, sorted=True)
    pred = pred.t()
    correct = pred.eq(y_true.view(1, -1).expand_as(pred))

    res: List[Tensor] = []
    for k in top_k:
        correct_k = correct[:k].contiguous().view(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


R = TypeVar("R", Tensor, Tuple[Tensor, Union[Tensor, slice]], covariant=True)


class Comparator(Protocol[R]):
    @abstractmethod
    def __call__(self, y_pred: Tensor, *, y_true: Tensor) -> R:
        """Compare.

        :param y_pred: Predicted labels or raw logits of a classifier.
        :param y_true: Ground truth (correct) labels.

        :returns: An element-wise comparison between ``y_pred`` and ``y_true`` or a subset of them;
            if the latter, the second element returned should be a mask indicating which samples
            comprise that subset.
        """
        ...


C = TypeVar("C", bound=Comparator)
C_co = TypeVar("C_co", bound=Comparator, covariant=True)


@torch.no_grad()  # pyright: ignore
def nanmax(x: Tensor) -> Tensor:
    return torch.amax(torch.nan_to_num(x, nan=-torch.inf))


@torch.no_grad()  # pyright: ignore
def nanmin(x: Tensor) -> Tensor:
    return torch.amin(torch.nan_to_num(x, nan=torch.inf))


@torch.no_grad()  # pyright: ignore
def pdist_1d(x: Tensor) -> Tensor:
    return (x - x.unsqueeze(-1)).abs().squeeze()


@torch.no_grad()  # pyright: ignore
def max_difference_1d(x: Tensor) -> Tensor:
    return nanmax(pdist_1d(x))


@torch.no_grad()  # pyright: ignore
def weighted_nanmean(x: Tensor, *, weights: Tensor) -> Tensor:
    if x.numel() != weights.numel():
        raise RuntimeError(
            "'weights' and the input tensor being weighted, 'x', must match in size."
        )
    if torch.any(weights < 0):
        raise RuntimeError("'weights' must contain only non-negative elements ")
    denom = weights.nansum()
    return (weights * x).nansum() / denom


class Aggregator(Enum):
    MIN = (nanmin,)
    "Aggregate by taking the minimum."
    MAX = (nanmax,)
    "Aggregate by taking the maximum."
    MEAN = (torch.nanmean,)
    "Aggregate by taking the mean."
    MEADIAN = (torch.nanmedian,)
    "Aggregate by taking the median."
    DIFF = (pdist_1d,)
    "Aggregate by taking the pairwise (absolute) differences."
    MAX_DIFF = (max_difference_1d,)
    "Aggregate by taking the maximum of the pairwise (absolute) differences."

    def __init__(self, fn: Callable[[Tensor], Tensor]) -> None:
        """
        Metric aggregator.

        :param fn: Aggregation function."""
        self.fn = fn

    @torch.no_grad()  # pyright: ignore
    def __call__(self, x: Tensor) -> Tensor:
        """Apply the aggregation function associated with the enum member to the input.

        :param x: Input to be aggregated.
        :returns: Aggregated input.
        """
        return self.fn(x)


@overload
def merge_indices(
    *indices: Tensor, return_cardinalities: Literal[True]
) -> Tuple[Tensor, List[int]]:
    ...


@overload
def merge_indices(*indices: Tensor, return_cardinalities: Literal[False] = ...) -> Tensor:
    ...


def merge_indices(
    *indices: Tensor, return_cardinalities: bool = False
) -> Union[Tensor, Tuple[Tensor, List[int]]]:
    """
    Bijectively merge a sequence of index tensors into a single index tensor, such that each
    combination of possible indices from across the elements in ``group_ids`` is assigned a unique
    index.

    :param indices: (Long-type) index tensors.
    :param return_cardinalities: if ``True``, return sizes.
    :returns: A merged index set which uniquely indexes each possible combination in indices.

    :raises TypeError: If any elemnts of ``indices`` do not have dtype ``torch.long``.
    """
    cards: Optional[List[int]] = None
    if return_cardinalities:
        cards = []
    group_ids_ls = list(indices)
    first_elem = group_ids_ls.pop().clone().squeeze()
    if first_elem.dtype != torch.long:
        raise TypeError("All index tensors must have dtype `torch.long'.")
    unique_vals, unique_inv = first_elem.unique(return_inverse=True)
    index_set = unique_inv
    if some(cards):
        cards.append(len(unique_vals))

    for elem in group_ids_ls:
        elem = elem.squeeze()
        if elem.dtype != torch.long:
            raise TypeError("All index tensors must have dtype `torch.long'.")
        unique_vals, inv_indices = elem.unique(return_inverse=True)
        card = int(len(unique_vals))
        if some(cards):
            cards.append(card)
        index_set *= card
        index_set += cast(Tensor, inv_indices)

    if cards is None:
        return index_set
    return index_set, cards


def nans(*sizes: int, device: Optional[torch.device] = None) -> Tensor:
    return torch.full(tuple(sizes), fill_value=torch.nan, device=device)


def nans_like(tensor: Tensor, *, device: Optional[torch.device] = None) -> Tensor:
    return torch.full_like(tensor, fill_value=torch.nan, device=device)


@torch.no_grad()  # pyright: ignore
def _apply_groupwise_metric(
    *group_ids: Tensor,
    comparator: Comparator,
    aggregator: Optional[Aggregator],
    y_pred: Tensor,
    y_true: Tensor,
) -> Tensor:
    """
    Computes a groupwise metric given a ``comparator`` and ``aggregator``.

    :param comparator: Function used to assess the correctness of ``y_pred`` with respect
        to ``y_true``. Should return a score for each sample.

    :param aggregator: Function with which to aggregate over the group-wise scores.
        If ``None`` then no aggregation will be applied and scores will be returned for each group.

    :param y_pred: Predictions to be scored
    :param y_true: Ground truth (correct) labels.
    :param group_ids: Ground truth labels indicating the group-membership of each sample.

    :returns: The score(s) as determined by the :attr:`comparator` and :attr:`aggregator`.

    :raises ValueError: If ``y_pred``, ``y_true``, and ``s`` do not match in size at dimension 0
        (the 'batch' dimension).
    """
    y_pred = y_pred.squeeze()
    y_true = y_true.squeeze()
    index_set: Optional[Tensor] = None

    if group_ids:
        group_ids_ls = list(group_ids)
        first_elem = group_ids_ls.pop().clone().squeeze()
        index_set = first_elem.unique(return_inverse=True)[1]
        assert isinstance(index_set, Tensor)  # pyright was complaining without this

        for elem in group_ids_ls:
            if len(y_pred) != len(y_true) != len(elem):
                raise ValueError(
                    "'y_pred', 'y_true', and all elements of 'group_ids' must match in size at"
                    " dimension 0."
                )
            elem = elem.squeeze()
            unique_vals, inv_indices = elem.unique(return_inverse=True)
            index_set *= len(unique_vals)
            index_set += cast(Tensor, inv_indices)

    res = comparator(y_pred=y_pred, y_true=y_true)
    if isinstance(res, tuple):
        comps, comp_mask = res
    else:
        comps, comp_mask = res, slice(None)

    if some(index_set):
        res = index_set.max()
        scores = torch.scatter_reduce(
            input=nans(int(index_set.max() + 1)),
            src=comps,
            dim=0,
            index=index_set[comp_mask],
            reduce="mean",
            include_self=False,
        )
        if some(aggregator):
            scores = aggregator(scores)
        return scores

    return comps.mean()


A = TypeVar("A", Aggregator, None)
A_co = TypeVar("A_co", Aggregator, None, covariant=True)


class Metric(Protocol[C_co, A_co]):
    @staticmethod
    def __call__(y_pred: Tensor, *, y_true: Tensor) -> Tensor:
        ...


class GroupwiseMetric(Protocol[C_co, A_co]):
    @staticmethod
    def __call__(y_pred: Tensor, *, y_true: Tensor, s: Tensor) -> Tensor:
        ...


def groupwise_metric(
    comparator: C, *, aggregator: A, cond_on_pred: bool = False
) -> GroupwiseMetric[C, A]:
    """Converts a given ``comparator`` and ``aggregator`` into a group-wise metric.

    :param comparator: Function used to assess the correctness of ``y_pred`` with respect
        to ``y_true``. Should return a score for each sample.
    :param aggregator: Function with which to aggregate over the group-wise scores.
        If ``None`` then no aggregation will be applied and scores will be returned for each group.
    :param cond_on_pred: Whethr to condition on predictions.

    :returns: A group-wise metric formed from ``comparator`` and ``aggregator``.
    """

    @wraps(comparator)
    def _decorated_comparator(y_pred: Tensor, *, y_true: Tensor, s: Tensor) -> Tensor:
        return _apply_groupwise_metric(
            s,
            y_pred if cond_on_pred else y_true,
            comparator=comparator,
            aggregator=aggregator,
            y_pred=y_pred,
            y_true=y_true,
        )

    return _decorated_comparator


def subclasswise_metric(
    comparator: C,
    *,
    aggregator: A,
) -> GroupwiseMetric[C, A]:
    """Converts a given ``comparator`` and ``aggregator`` into a subclass-wise metric.

    :param comparator: Function used to assess the correctness of ``y_pred`` with respect
        to ``y_true``. Should return a score for each sample.

    :param aggregator: Function with which to aggregate over the subclass-wise scores.
        If ``None`` then no aggregation will be applied and scores will be returned for each group.

    :returns: A subclass-wise metric formed from ``comparator`` and ``aggregator``.
    """

    @wraps(comparator)
    def _decorated_comparator(y_pred: Tensor, *, y_true: Tensor, s: Tensor) -> Tensor:
        return _apply_groupwise_metric(
            s, comparator=comparator, aggregator=aggregator, y_pred=y_pred, y_true=y_true
        )

    return _decorated_comparator


def classwise_metric(
    comparator: C,
    *,
    aggregator: A,
    cond_on_pred: bool = False,
) -> Metric[C, A]:
    """Converts a given ``comparator`` and ``aggregator`` into a subclass-wise metric.

    :param comparator: Function used to assess the correctness of ``y_pred`` with respect
        to ``y_true``. Should return a score for each sample.

    :param aggregator: Function with which to aggregate over the subclass-wise scores.
        If ``None`` then no aggregation will be applied and scores will be returned for each group.
    :param cond_on_pred: Whethr to condition on predictions.

    :returns: A subclass-wise metric formed from ``comparator`` and ``aggregator``.
    """

    @wraps(comparator)
    def _decorated_comparator(y_pred: Tensor, *, y_true: Tensor) -> Tensor:
        return _apply_groupwise_metric(
            y_pred if cond_on_pred else y_true,
            comparator=comparator,
            aggregator=aggregator,
            y_pred=y_pred,
            y_true=y_true,
        )

    return _decorated_comparator


def equal(y_pred: Tensor, *, y_true: Tensor) -> Tensor:
    y_true = torch.atleast_1d(y_true.squeeze()).long()
    if len(y_pred) != len(y_true):
        raise ValueError("'y_pred' and 'y_true' must match in size at dimension 0.")
    # Interpret floating point predictions as potential logits and attempt to convert
    # them to hard predictions.
    if y_pred.is_floating_point():
        y_pred = hard_prediction(y_pred)
    return (y_pred == y_true).float()


def conditional_equal(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    y_pred_cond: Optional[int] = None,
    y_true_cond: Optional[int] = None,
) -> Tuple[Tensor, Tensor]:
    mask = torch.ones_like(y_pred, dtype=torch.bool)
    if some(y_pred_cond):
        mask &= y_pred == y_pred_cond
    if some(y_true_cond):
        mask &= y_true == y_true_cond
    comps = equal(y_pred=y_pred[mask], y_true=y_true[mask])
    return comps, mask


@torch.no_grad()  # pyright: ignore
def accuracy(y_pred: Tensor, *, y_true: Tensor) -> Tensor:
    """Computes the (aggregate) accuracy given predicted and ground-truth labels.

    :param y_pred: Predictions to be scored.
    :param y_true: Ground truth (correct) labels.

    :returns: Aggregate accuracy.
    """
    return Aggregator.MEAN(equal(y_pred=y_pred.squeeze(), y_true=y_true.squeeze()))


robust_accuracy = subclasswise_metric(comparator=equal, aggregator=Aggregator.MIN)
accuracy_per_subclass = subclasswise_metric(comparator=equal, aggregator=None)
subclass_balanced_accuracy = subclasswise_metric(comparator=equal, aggregator=Aggregator.MEAN)
robust_gap = subclasswise_metric(comparator=equal, aggregator=Aggregator.MAX_DIFF)

group_balanced_accuracy = groupwise_metric(comparator=equal, aggregator=Aggregator.MEAN)
accuracy_per_group = groupwise_metric(comparator=equal, aggregator=None)
# Computes the TPR for each classs (in a one-vs-rest fashion) and subclass and returns the minimum
# over the resulting set.
robust_ovr_tpr = groupwise_metric(comparator=equal, aggregator=Aggregator.MIN)

accuracy_per_class = classwise_metric(comparator=equal, aggregator=None)
balanced_accuracy = classwise_metric(
    comparator=equal, aggregator=Aggregator.MEAN, cond_on_pred=False
)
precision_per_class = classwise_metric(comparator=equal, aggregator=None, cond_on_pred=True)
precision_per_subclass = groupwise_metric(comparator=equal, aggregator=None, cond_on_pred=True)

balanced_precision = classwise_metric(comparator=equal, aggregator=None, cond_on_pred=True)

tpr_per_subclass = subclasswise_metric(
    comparator=partial(conditional_equal, y_true_cond=1), aggregator=None
)
tnr_per_subclass = subclasswise_metric(
    comparator=partial(conditional_equal, y_true_cond=0), aggregator=None
)
tpr_differences = subclasswise_metric(
    comparator=partial(conditional_equal, y_true_cond=1), aggregator=Aggregator.DIFF
)
tnr_differences = subclasswise_metric(
    comparator=partial(conditional_equal, y_true_cond=0), aggregator=Aggregator.DIFF
)
robust_tpr = subclasswise_metric(
    comparator=partial(conditional_equal, y_true_cond=1), aggregator=Aggregator.MIN
)
robust_tnr = subclasswise_metric(
    comparator=partial(conditional_equal, y_true_cond=0), aggregator=Aggregator.MIN
)


@torch.no_grad()  # pyright: ignore
def pad_to_size(size: int, *, src: Tensor, index: Tensor, value: float = 0.0) -> Tensor:
    padding = src.new_full((size,), fill_value=value)
    return padding.scatter_(dim=0, src=src, index=index)


def fscore(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    s: Optional[Tensor] = None,
    beta: float = 1.0,
    inner_summand: Optional[Literal["y_true", "s"]] = None,
    ignore_unsupported: bool = True,
    aggregator: Optional[Aggregator] = None,
) -> Tensor:
    """Computes F-beta score between ``y_pred`` and ``y_true`` with optional subclass-conditioning.

    :param y_pred: Predicted labels.
    :param y_true: Target labels.
    :param s: Subclass labels.
    :param beta: Beta coefficient that determines the weight of recall relative to precision.
        ``beta < 1`` lends more weight to precision, while ``beta > 1`` favors recal
    :param inner_summand: Which conditioning factor, if any, to sum over prior to the final
        aggregation when conditioning on the subclass labels.
    :param ignore_unsupported: Whether to ignore ``y_pred`` values that are unsupported by
        ``y_true`` (labels appearing in the predicted but not ground-truth labels). Note that
        class:`sklearn.metrics.f1_score` does *not* do this and instead includes unsupported values
        in its aggregation, i.e. classes included only in ``y_pred`` are included in the
        class-conditional F-scores as 0 components that influence the aggregation step.
    :param aggregator: Function with which to aggregate over the scores.
        If ``None`` then no aggregation will be applied and scores will be returned for each group.

    :returns: The (optionally aggregated) F-beta score given predictions ``y_pred`` and targets
        ``y_pred``.
    """
    # map the predicted and ground-truth labels to a common basis
    y_unique, rel_indices = torch.unique(torch.cat((y_true, y_pred)), return_inverse=True)
    y_true, y_pred = rel_indices.chunk(2)
    card_y = len(y_unique)

    if s is None:
        rec_ids = y_true
        rec_ids_u = rec_ids.unique()
        card_s = None
        prec_ids = y_pred
        target_size = card_y
    else:
        s_u, s_inv = s.unique(return_inverse=True)
        card_s = len(s_u)
        prec_ids = y_pred * card_s + s_inv
        rec_ids = y_true * card_s + s_inv
        rec_ids_u = rec_ids.unique()
        target_size = card_y * card_s

    precs = _apply_groupwise_metric(
        prec_ids,
        comparator=equal,
        y_pred=y_pred,
        y_true=y_true,
        aggregator=None,
    )
    recs = _apply_groupwise_metric(
        rec_ids,
        comparator=equal,
        y_pred=y_pred,
        y_true=y_true,
        aggregator=None,
    )

    if len(precs) < target_size:
        precs = pad_to_size(size=target_size, index=prec_ids.unique(), src=precs, value=torch.nan)
    if len(recs) < target_size:
        recs = pad_to_size(size=target_size, index=rec_ids_u, src=recs, value=torch.nan)

    beta_sq = beta**2
    f1s = (1 + beta_sq) * (precs * recs) / (beta_sq * precs + recs)
    if ignore_unsupported:
        # Ignore predictions that are unsupported by the ground-truth label set, supp(``y_true``).
        f1s = f1s.gather(0, rec_ids_u)
    f1s = f1s.nan_to_num_(nan=0.0)

    if some(inner_summand) and some(s):
        card_s = len(torch.unique(s))
        reduction_dim = int(inner_summand == "s")
        f1s = f1s.view(-1, card_s).nanmean(reduction_dim)

    if aggregator is None:
        return f1s
    return aggregator(f1s)


def robust_fscore(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    s: Tensor,
    beta: float = 1.0,
    ignore_unsupported: bool = True,
) -> Tensor:
    return fscore(
        y_pred=y_pred,
        y_true=y_true,
        s=s,
        beta=beta,
        inner_summand="y_true",
        ignore_unsupported=ignore_unsupported,
        aggregator=Aggregator.MIN,
    )


def fscore_per_class(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    beta: float = 1.0,
    ignore_unsupported: bool = True,
) -> Tensor:
    return fscore(
        y_pred=y_pred,
        y_true=y_true,
        beta=beta,
        inner_summand=None,
        ignore_unsupported=ignore_unsupported,
        aggregator=None,
    )


def fscore_per_subclass(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    s: Tensor,
    beta: float = 1.0,
    ignore_unsupported: bool = True,
) -> Tensor:
    return fscore(
        y_pred=y_pred,
        y_true=y_true,
        s=s,
        beta=beta,
        inner_summand="y_true",
        ignore_unsupported=ignore_unsupported,
        aggregator=None,
    )


def macro_fscore(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    beta: float = 1.0,
    ignore_unsupported: bool = True,
) -> Tensor:
    return fscore(
        y_pred=y_pred,
        y_true=y_true,
        s=None,
        beta=beta,
        inner_summand=None,
        ignore_unsupported=ignore_unsupported,
        aggregator=Aggregator.MEAN,
    )


def fscore_per_group(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    s: Tensor,
    beta: float = 1.0,
    ignore_unsupported: bool = True,
) -> Tensor:
    return fscore(
        y_pred=y_pred,
        y_true=y_true,
        s=s,
        beta=beta,
        inner_summand=None,
        ignore_unsupported=ignore_unsupported,
        aggregator=None,
    )


def robust_fscore_gap(
    y_pred: Tensor,
    *,
    y_true: Tensor,
    s: Tensor,
    beta: float = 1.0,
    ignore_unsupported: bool = True,
) -> Tensor:
    return fscore(
        y_pred=y_pred,
        y_true=y_true,
        s=s,
        beta=beta,
        inner_summand="y_true",
        ignore_unsupported=ignore_unsupported,
        aggregator=Aggregator.MAX_DIFF,
    )



from typing import Optional
from typing_extensions import TypeAlias, override

import numpy as np
import numpy.typing as npt
import torch
from torch import Tensor

from conduit.data.structures import TernarySample
from conduit.types import IndexType

from .base import CdtVisionDataset

SampleType: TypeAlias = TernarySample


class DummyVisionDataset(CdtVisionDataset[SampleType, Tensor, Tensor]):
    def __init__(
        self,
        channels: int,
        height: int,
        width: int,
        s_card: Optional[int] = 2,
        y_card: Optional[int] = 2,
        num_samples: int = 10_000,
    ) -> None:
        self.channels = channels
        self.height = height
        self.width = width
        s = torch.randint(s_card, (num_samples,)) if s_card is not None else None
        y = torch.randint(y_card, (num_samples,)) if y_card is not None else None
        x = np.array([""] * num_samples)
        super().__init__(x=x, s=s, y=y, image_dir="")

    @override
    def _load_image(self, index: IndexType) -> npt.NDArray[np.uint8]:
        return np.random.randint(
            0, 256, size=(self.height, self.width, self.channels), dtype="uint8"
        )



"""Base class for vision datasets."""

from abc import abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union, final
from typing_extensions import override

import albumentations as A  # type: ignore
from albumentations.pytorch import ToTensorV2  # type: ignore
from torch import Tensor

from conduit.data.constants import IMAGENET_STATS
from conduit.data.datamodules.base import CdtDataModule
from conduit.data.datasets.base import I
from conduit.data.datasets.vision import AlbumentationsTform, ImageTform, ImageTransformer
from conduit.data.datasets.vision.base import CdtVisionDataset
from conduit.data.structures import ImageSize, MeanStd, TrainValTestSplit

__all__ = ["CdtVisionDataModule"]


@dataclass(kw_only=True)
class CdtVisionDataModule(CdtDataModule[ImageTransformer, I]):
    root: Union[str, Path] = field(kw_only=False)
    train_tf: Optional[ImageTform] = None
    test_tf: Optional[ImageTform] = None
    norm_values: Optional[MeanStd] = field(default=IMAGENET_STATS, init=False)

    @property
    @final
    def train_transforms(self) -> ImageTform:
        return self._default_train_transforms if self.train_tf is None else self.train_tf

    @train_transforms.setter
    def train_transforms(self, transform: Optional[ImageTform]) -> None:
        self.train_tf = transform
        if isinstance(self._train_data, ImageTransformer):
            self._train_data.transform = transform

    @property
    @final
    def test_transforms(self) -> ImageTform:
        return self._default_test_transforms if self.test_tf is None else self.test_tf

    @test_transforms.setter
    @final
    def test_transforms(self, transform: Optional[ImageTform]) -> None:
        self.test_tf = transform
        if isinstance(self._val_data, ImageTransformer):
            self._val_data.transform = transform
        if isinstance(self._test_data, ImageTransformer):
            self._test_data.transform = transform

    @property
    def _default_train_transforms(self) -> A.Compose:
        transform_ls: List[AlbumentationsTform] = [A.ToFloat()]
        if self.norm_values is not None:
            # `max_pixel_value` has to be 1.0 here because of `ToFloat()`
            transform_ls.append(
                A.Normalize(
                    mean=self.norm_values.mean,
                    std=self.norm_values.std,
                    max_pixel_value=1.0,
                )
            )
        transform_ls.append(ToTensorV2())
        return A.Compose(transform_ls)

    @property
    @override
    def dim_x(self) -> ImageSize:
        """
        Returns the dimensions of the first input (x).

        :returns: ImageSize object containing the dimensions (C, H, W) of the (first) input.
        """
        return ImageSize(*super().dim_x)

    @override
    def size(self) -> ImageSize:
        """Alias for ``dim_x``.

        :returns: ImageSize object containing the dimensions (C, H, W) of the (first) input.
        """
        return self.dim_x

    @property
    def _default_test_transforms(self) -> A.Compose:
        transform_ls: List[AlbumentationsTform] = [A.ToFloat()]
        if self.norm_values is not None:
            transform_ls.append(A.Normalize(mean=self.norm_values.mean, std=self.norm_values.std))
        transform_ls.append(ToTensorV2())
        return A.Compose(transform_ls)

    @abstractmethod
    def _get_image_splits(self) -> TrainValTestSplit[CdtVisionDataset[I, Tensor, Tensor]]:
        raise NotImplementedError()

    @final
    @override
    def _get_splits(self) -> TrainValTestSplit[ImageTransformer]:
        train, val, test = self._get_image_splits()
        return TrainValTestSplit(
            train=ImageTransformer(train, transform=self.train_transforms),
            val=ImageTransformer(val, transform=self.test_transforms),
            test=ImageTransformer(test, transform=self.test_transforms),
        )



"""NICO data-module."""

from dataclasses import dataclass
from typing import Any, Optional
from typing_extensions import override

import albumentations as A  # type: ignore

from conduit.data.datamodules.vision.base import CdtVisionDataModule
from conduit.data.datasets.utils import stratified_split
from conduit.data.datasets.vision.nico import NICO, NicoSuperclass, SampleType
from conduit.data.structures import TrainValTestSplit

__all__ = ["NICODataModule"]


@dataclass(kw_only=True)
class NICODataModule(CdtVisionDataModule[SampleType]):
    """Data-module for the NICO dataset."""

    image_size: int = 224
    class_train_props: Optional[dict] = None
    superclass: NicoSuperclass = NicoSuperclass.ANIMALS

    @property
    @override
    def _default_train_transforms(self) -> A.Compose:
        base_transforms = A.Compose(
            [
                A.Resize(self.image_size, self.image_size),
                A.CenterCrop(self.image_size, self.image_size),
            ]
        )
        normalization = super()._default_train_transforms
        return A.Compose([base_transforms, normalization])

    @property
    @override
    def _default_test_transforms(self) -> A.Compose:
        return self._default_train_transforms

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        NICO(root=self.root, download=True)

    @override
    def _get_image_splits(self) -> TrainValTestSplit[NICO]:
        all_data = NICO(root=self.root, superclass=self.superclass, transform=None)
        train_val_prop = 1 - self.test_prop
        train_val_data, test_data = stratified_split(
            all_data,
            default_train_prop=train_val_prop,
            train_props=self.class_train_props,
            seed=self.seed,
        )
        val_data, train_data = train_val_data.random_split(
            props=self.val_prop / train_val_prop, seed=self.seed
        )

        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



"""Crime Dataset."""

from dataclasses import dataclass

from ethicml.data import Crime, Dataset
from ethicml.data import CrimeSplits as CrimeSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["CrimeDataModule", "CrimeSens"]


@dataclass(kw_only=True)
class CrimeDataModule(EthicMlDataModule):
    """Data Module for the Crime Dataset."""

    sens_feat: CrimeSens = CrimeSens.RACE_BINARY
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Crime(
            split=self.sens_feat, discrete_only=self.disc_feats_only, invert_s=self.invert_s
        )



"""Waterbirds data-module."""

from dataclasses import dataclass
from typing import Any
from typing_extensions import override

import albumentations as A  # type: ignore

from conduit.data.datamodules.vision.base import CdtVisionDataModule
from conduit.data.datasets.vision.waterbirds import Waterbirds
from conduit.data.structures import TrainValTestSplit

__all__ = ["WaterbirdsDataModule"]


@dataclass(kw_only=True)
class WaterbirdsDataModule(CdtVisionDataModule[Waterbirds.SampleType]):
    """Data-module for the Waterbirds dataset."""

    image_size: int = 224
    use_predefined_splits: bool = False

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        Waterbirds(root=self.root, download=True)

    @property
    @override
    def _default_train_transforms(self) -> A.Compose:
        # We use the transoform pipeline described in https://arxiv.org/abs/2008.06775
        # rather than that described in the paper in which the dataset was first introduced
        # (Sagawa et al); these differ in in the respect that the latter center-crops the images
        # before resizing - not doing so makes the task more difficult due to the background
        # (serving as a spurious attribute) being more prominent.
        base_transforms = A.Compose(
            [
                A.Resize(self.image_size, self.image_size),
            ]
        )
        normalization = super()._default_train_transforms
        return A.Compose([base_transforms, normalization])

    @property
    @override
    def _default_test_transforms(self) -> A.Compose:
        return self._default_train_transforms

    @override
    def _get_image_splits(self) -> TrainValTestSplit[Waterbirds]:
        # Split the data according to the pre-defined split indices
        if self.use_predefined_splits:
            train_data, val_data, test_data = (
                Waterbirds(root=self.root, split=split) for split in Waterbirds.Split
            )
        # Split the data randomly according to test- and val-prop
        else:
            all_data = Waterbirds(root=self.root, transform=None)
            val_data, test_data, train_data = all_data.random_split(
                props=(self.val_prop, self.test_prop), seed=self.seed
            )
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



"""Base class for audio datasets."""

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Optional, final
from typing_extensions import override

from torch import Tensor
import torchaudio.transforms as T  # type: ignore

from conduit.data.datamodules.base import CdtDataModule
from conduit.data.datasets.audio.base import CdtAudioDataset
from conduit.data.datasets.base import I
from conduit.data.datasets.utils import AudioTform
from conduit.data.datasets.wrappers import AudioTransformer
from conduit.data.structures import TrainValTestSplit

__all__ = ["CdtAudioDataModule"]


@dataclass(kw_only=True)
class CdtAudioDataModule(CdtDataModule[AudioTransformer, I]):
    root: str = field(kw_only=False)
    train_tf: Optional[AudioTform] = None
    test_tf: Optional[AudioTform] = None

    @property
    @final
    def train_transforms(self) -> AudioTform:
        return self._default_train_transforms if self.train_tf is None else self.train_tf

    @train_transforms.setter
    def train_transforms(self, transform: Optional[AudioTform]) -> None:
        self.train_tf = transform
        if isinstance(self._train_data, AudioTransformer):
            self._train_data.transform = transform

    @property
    @final
    def test_transforms(self) -> AudioTform:
        return self._default_test_transforms if self.test_tf is None else self.test_tf

    @test_transforms.setter
    @final
    def test_transforms(self, transform: Optional[AudioTform]) -> None:
        self.test_tf = transform
        if isinstance(self._val_data, AudioTransformer):
            self._val_data.transform = transform
        if isinstance(self._test_data, AudioTransformer):
            self._test_data.transform = transform

    @property
    def _default_train_transforms(self) -> T.Spectrogram:
        return T.Spectrogram()

    @property
    def _default_test_transforms(self) -> T.Spectrogram:
        return T.Spectrogram()

    @abstractmethod
    def _get_audio_splits(self) -> TrainValTestSplit[CdtAudioDataset[I, Tensor, Tensor]]:
        raise NotImplementedError()

    @override
    def _get_splits(self) -> TrainValTestSplit[AudioTransformer]:
        train, val, test = self._get_audio_splits()
        return TrainValTestSplit(
            train=AudioTransformer(train, transform=self.train_transforms),
            val=AudioTransformer(val, transform=self.test_transforms),
            test=AudioTransformer(test, transform=self.test_transforms),
        )



from dataclasses import dataclass
from typing import Generic, List, Optional, Sequence, Tuple, TypeVar, Union, overload
from typing_extensions import Self, override

from ranzen.misc import gcopy
import torch
from torch import Tensor
import torchvision.transforms as T  # type: ignore
import torchvision.transforms.functional as TF  # type: ignore

from conduit.data.constants import IMAGENET_STATS
from conduit.data.datasets.vision.utils import (
    ImageTform,
    PillowTform,
    apply_image_transform,
    img_to_tensor,
)
from conduit.data.structures import InputContainer, MeanStd, RawImage, concatenate_inputs

from .image import RandomGaussianBlur, RandomSolarize

__all__ = ["MultiCropOutput", "MultiCropTransform", "MultiViewPair"]


@dataclass
class MultiViewPair(InputContainer):
    v1: Tensor
    v2: Tensor

    def __post_init__(self) -> None:
        if self.v1.size() != self.v2.size():
            raise AttributeError("'v1' and 'v2' must have the same shape.")

    @override
    def __len__(self) -> int:
        return len(self.v1)

    @override
    def __add__(self, other: Self) -> Self:
        copy = gcopy(self, deep=False)
        is_batched = self.v1.ndim == 4
        copy.v1 = concatenate_inputs(copy.v1, other.v1, is_batched=is_batched)
        copy.v2 = concatenate_inputs(copy.v2, other.v2, is_batched=is_batched)

        return copy

    def size(self) -> torch.Size:
        return self.v1.size()

    @property
    def shape(self) -> torch.Size:
        return self.v1.shape

    def merge(self) -> Tensor:
        is_batched = self.v1.ndim == 4
        return concatenate_inputs(self.v1, self.v2, is_batched=is_batched)

    @property
    def anchor(self) -> Tensor:
        return self.v1

    @property
    def target(self) -> Tensor:
        return self.v2

    @property
    def num_sources(self) -> int:
        return len(self)


@dataclass
class MultiCropOutput(InputContainer):
    global_views: MultiViewPair
    local_views: Tensor

    @property
    def num_sources(self) -> int:
        """The number of samples from which the views were generated."""
        return len(self.global_views)

    @property
    def num_global_crops(self) -> int:
        return 2

    @property
    def num_local_crops(self) -> int:
        if self.local_views is None:
            return 0
        return len(self.local_views) // len(self.global_views)

    @property
    def num_crops(self) -> int:
        return self.num_global_crops + self.num_local_crops

    @property
    def global_crop_size(self) -> Tuple[int, int, int]:
        return (self.global_views.shape[1], self.global_views.shape[2], self.global_views.shape[3])

    @property
    def local_crop_size(self) -> Tuple[int, int, int]:
        if self.local_views is None:
            raise AttributeError("Cannot retrieve the local-crop size as 'local_' is 'None'.")
        return (self.local_views.shape[1], self.local_views.shape[2], self.local_views.shape[3])

    @property
    def shape(self) -> torch.Size:
        """Shape of the global crops."""
        return self.global_views.shape

    def astuple(self) -> Tuple[Tensor, Tensor]:
        return (self.global_views.merge(), self.local_views)

    @property
    def anchor(self) -> Tuple[Tensor, Tensor]:
        return (self.global_views.v1, self.local_views)

    @property
    def target(self) -> Tensor:
        return self.global_views.v2

    @override
    def __len__(self) -> int:
        """Total number of crops."""
        return len(self.global_views) + len(self.local_views)

    @override
    def __add__(self, other: Self) -> Self:
        copy = gcopy(self, deep=False)
        copy.global_views = copy.global_views + other.global_views
        if copy.local_views is None:
            if other.local_views is not None:
                copy.local_views = other.local_views
        else:
            if other.local_views is not None:
                copy.local_views = copy.local_views + other.local_views
                is_batched = copy.local_views.ndim == 4
                copy.local_views = concatenate_inputs(
                    copy.local_views, other.local_views, is_batched=is_batched
                )
        return copy


LT = TypeVar("LT", bound=Optional[ImageTform])


class MultiCropTransform(Generic[LT]):
    def __init__(
        self,
        *,
        global_transform_1: ImageTform,
        global_transform_2: Optional[ImageTform] = None,
        local_transform: LT = None,
        local_crops_number: int = 8,
    ) -> None:
        self.global_transform_1 = global_transform_1
        self.global_transform_2 = (
            global_transform_1 if global_transform_2 is None else global_transform_2
        )
        if (local_transform is not None) and (local_crops_number <= 0):
            raise AttributeError(
                " 'local_crops' must be a positive integer if 'local_transform' is defined."
            )
        self.local_transform = local_transform
        self.local_crops_number = local_crops_number

    @staticmethod
    def _apply_transform(image: RawImage, transform: ImageTform) -> Tensor:
        view = apply_image_transform(image, transform=transform)
        if not isinstance(view, Tensor):
            view = img_to_tensor(view)
        return view

    @overload
    def __call__(self: "MultiCropTransform[ImageTform]", image: RawImage) -> MultiCropOutput:
        ...

    @overload
    def __call__(self: "MultiCropTransform[None]", image: RawImage) -> MultiViewPair:
        ...

    def __call__(
        self: "MultiCropTransform", image: RawImage
    ) -> Union[MultiCropOutput, MultiViewPair]:
        global_crop_v1 = self._apply_transform(image, transform=self.global_transform_1)
        global_crop_v2 = self._apply_transform(image, transform=self.global_transform_2)
        gc_pair = MultiViewPair(v1=global_crop_v1, v2=global_crop_v2)

        if (self.local_transform is None) or (self.local_crops_number == 0):
            return gc_pair
        local_crops = torch.stack(
            [
                self._apply_transform(image, transform=self.local_transform)
                for _ in range(self.local_crops_number)
            ],
            dim=0,
        )

        return MultiCropOutput(global_views=gc_pair, local_views=local_crops)

    @classmethod
    def with_dino_transform(
        cls,
        *,
        global_crop_size: Union[int, Sequence[int]] = 224,
        local_crop_size: Union[int, Sequence[int]] = 96,
        norm_values: Optional[MeanStd] = IMAGENET_STATS,
        global_crops_scale: Tuple[float, float] = (0.4, 1.0),
        local_crops_scale: Tuple[float, float] = (0.05, 0.4),
        local_crops_number: int = 8,
    ) -> "MultiCropTransform":
        flip_and_color_jitter = T.Compose(
            [
                T.RandomHorizontalFlip(p=0.5),
                T.RandomApply(
                    [T.ColorJitter(brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1)], p=0.8
                ),
                T.RandomGrayscale(p=0.2),
            ]
        )
        normalize_ls: List[PillowTform] = [T.ToTensor()]
        if norm_values is not None:
            normalize_ls.append(
                T.Normalize(mean=norm_values.mean, std=norm_values.std),
            )
        normalize = T.Compose(normalize_ls)

        # first global crop
        global_transform_1 = T.Compose(
            [
                T.RandomResizedCrop(
                    global_crop_size,
                    scale=global_crops_scale,
                    interpolation=TF.InterpolationMode.BICUBIC,
                ),
                flip_and_color_jitter,
                RandomGaussianBlur(1.0),
                normalize,
            ]
        )
        # second global crop
        global_transform_2 = T.Compose(
            [
                T.RandomResizedCrop(
                    global_crop_size,
                    scale=global_crops_scale,
                    interpolation=TF.InterpolationMode.BICUBIC,
                ),
                flip_and_color_jitter,
                RandomGaussianBlur(0.1),
                RandomSolarize(0.2),
                normalize,
            ]
        )
        # transformation for the local small crops
        local_transform = None
        if local_crops_number > 0:
            local_transform = T.Compose(
                [
                    T.RandomResizedCrop(
                        local_crop_size,
                        scale=local_crops_scale,
                        interpolation=TF.InterpolationMode.BICUBIC,
                    ),
                    flip_and_color_jitter,
                    RandomGaussianBlur(p=0.5),
                    normalize,
                ]
            )

        return MultiCropTransform(
            global_transform_1=global_transform_1,
            global_transform_2=global_transform_2,
            local_transform=local_transform,
            local_crops_number=local_crops_number,
        )



from typing import Optional
from typing_extensions import Self, override

import torch
from torch import Tensor, nn
import torch.nn.functional as F

from conduit.types import IndexType, Indexable, Sized

__all__ = ["MemoryBank"]


@torch.no_grad()  # pyright: ignore
def l2_hypersphere_init(capacity: int, *, dim: int) -> Tensor:
    return F.normalize(torch.randn(capacity, dim), dim=1)


class MemoryBank(nn.Module, Indexable, Sized):
    """Fixed-sized Memory Bank implementing a FIFO-update rule."""

    memory: Tensor

    def __init__(self, memory: Tensor) -> None:
        super().__init__()
        self.capacity = memory.size(0)
        self.dim = memory.shape[1:]
        self.register_buffer(name="memory", tensor=memory)
        self._ptr_pos = 0

    @classmethod
    @torch.no_grad()  # pyright: ignore
    def with_l2_hypersphere_init(cls, capacity: int, *, dim: int) -> Self:
        return cls(l2_hypersphere_init(capacity=capacity, dim=dim))

    @classmethod
    @torch.no_grad()  # pyright: ignore
    def with_randint_init(cls, capacity: int, *, dim: int, high: int, low: int = 0) -> Self:
        return cls(torch.randint(low=low, high=high, size=(capacity, dim)))

    @classmethod
    @torch.no_grad()  # pyright: ignore
    def with_constant_init(
        cls, capacity: int, *, dim: int, value: float, dtype: Optional[torch.dtype] = None
    ) -> Self:
        return cls(torch.full(size=(capacity, dim), fill_value=value, dtype=dtype))

    def clone(self) -> Tensor:
        return self.memory.clone()

    @torch.no_grad()  # pyright: ignore
    def push(self, values: Tensor) -> None:
        values = values.detach()
        num_values = values.size(0)
        # Coerce the values into being 2-dimensional.
        values = values.view(num_values, -1)
        residual = self.capacity - self._ptr_pos
        # Advance the pointer by 'num_values' steps.
        new_ptr_pos = (self._ptr_pos + num_values) % self.capacity

        if residual <= num_values:
            self.memory[self._ptr_pos :] = values[:residual]
            self.memory[:new_ptr_pos] = values[residual:]
        else:
            self.memory[self._ptr_pos : new_ptr_pos] = values
        self._ptr_pos = new_ptr_pos

    @override
    def __getitem__(self, index: IndexType) -> Tensor:
        return self.memory[index]

    @override
    def __len__(self) -> int:
        return len(self.memory)



from .base import *
from .camelyon17 import *
from .celeba import *
from .cmnist import *
from .fmow import *
from .isic import *
from .iwildcam import *
from .nico import *
from .nico_plus_plus import *
from .nih import *
from .pacs import *
from .ssrp import *
from .utils import *
from .waterbirds import *
from .wrappers import *



"""PACS Dataset."""

from enum import auto
from pathlib import Path
from typing import ClassVar, List, Optional, Union
from typing_extensions import Self, TypeAlias

import numpy as np
import pandas as pd
from ranzen import StrEnum, str_to_enum
import torch
from torch import Tensor

from conduit.data.datasets.utils import GdriveFileInfo, download_from_gdrive
from conduit.data.structures import TernarySample, TrainTestSplit

from .base import CdtVisionDataset
from .utils import ImageTform

__all__ = [
    "PACS",
    "PacsDomain",
]


SampleType: TypeAlias = TernarySample


class PacsDomain(StrEnum):
    PHOTO = auto()
    ART_PAINTING = auto()
    CARTOON = auto()
    SKETCH = auto()


class PACS(CdtVisionDataset[TernarySample, Tensor, Tensor]):
    """
    PACS (photo (P), art painting (A), cartoon (C), and sketch (S)) dataset for evaluating domain
    generalisation as introduced in `PACS`_ 'Deeper, Broader and Artier Domain Generalization'.

    .. _PACS:
        https://arxiv.org/abs/1710.03077
    """

    SampleType: TypeAlias = TernarySample
    Domain: TypeAlias = PacsDomain

    _FILE_INFO: ClassVar[GdriveFileInfo] = GdriveFileInfo(
        name="PACS.zip",
        id="1lNOnDplyxhdDdyfb4KRIh0p0uom8aS1i",
        md5="5e43a6e01e53567923621ae5ce025f4e",
    )

    def __init__(
        self,
        root: Union[str, Path],
        *,
        download: bool = True,
        transform: Optional[ImageTform] = None,
        domains: Optional[Union[Domain, str, List[Union[str, Domain]]]] = None,
    ) -> None:
        self.domains: Union[None, PacsDomain, List[PacsDomain]]
        if isinstance(domains, str):
            self.domains = PacsDomain(domains)
        elif isinstance(domains, list):
            domains_ = [PacsDomain(elem) for elem in domains]
            self.domains = domains_
        else:
            self.domains = domains

        self.root = Path(root)
        self.download = download
        self._base_dir = self.root / self.__class__.__name__
        self._metadata_path = self._base_dir / "metadata.csv"

        if self.download:
            download_from_gdrive(file_info=self._FILE_INFO, root=self.root, logger=self.logger)
        elif not self._check_unzipped():
            raise FileNotFoundError(
                f"Data not found at location {self._base_dir.resolve()}. Have you downloaded it?"
            )
        if not self._metadata_path.exists():
            self._extract_metadata()

        self.metadata = pd.read_csv(self._base_dir / "metadata.csv")

        # Label decoders (for mapping from the label encodings back to their original string
        # values).
        self.domain_label_decoder = (
            self.metadata[["domain", "domain_le"]].set_index("domain_le").to_dict()["domain"]
        )
        self.class_label_decoder = (
            self.metadata[["class", "class_le"]].set_index("class_le").to_dict()["class"]
        )

        if isinstance(self.domains, PacsDomain):
            self.metadata = self.metadata[self.metadata["domain"] == str(self.domains)]
        elif isinstance(self.domains, list):
            str_list = [str(elem) for elem in self.domains]
            self.metadata = self.metadata[self.metadata["domain"].isin(str_list)]

        # # Divide up the dataframe into its constituent arrays because indexing with pandas is
        # # substantially slower than indexing with numpy/torch
        x = self.metadata["filepath"].to_numpy()
        y = torch.as_tensor(self.metadata["class_le"].to_numpy(), dtype=torch.long)
        s = torch.as_tensor(self.metadata["domain_le"].to_numpy(), dtype=torch.long)

        super().__init__(x=x, y=y, s=s, transform=transform, image_dir=self._base_dir)

    def _check_unzipped(self) -> bool:
        return all((self._base_dir / str(sc)).exists() for sc in PacsDomain)

    def _extract_metadata(self) -> None:
        """Extract domain/class information from the image filepaths and it save to csv."""
        self.logger.info("Extracting metadata.")
        image_paths: List[Path] = []
        for ext in ("jpg", "jpeg", "png"):
            image_paths.extend(self._base_dir.glob(f"**/*.{ext}"))
        image_paths_str = [str(image.relative_to(self._base_dir)) for image in image_paths]
        filepaths = pd.Series(image_paths_str)
        metadata = filepaths.str.split("/", expand=True).rename(
            columns={0: "domain", 1: "class", 2: "filename"}
        )
        metadata["filepath"] = filepaths
        metadata.sort_index(axis=1, inplace=True)
        metadata.sort_values(by=["filepath"], axis=0, inplace=True)
        metadata = self._label_encode_metadata(metadata)
        metadata.to_csv(self._metadata_path)

    @staticmethod
    def _label_encode_metadata(metadata: pd.DataFrame) -> pd.DataFrame:
        """Label encode the extracted concept/context/superclass information."""
        for col in metadata.columns:
            # Skip over filepath and filename columns
            if "file" not in col:
                # Add a new column containing the label-encoded data
                metadata[f"{col}_le"] = metadata[col].factorize()[0]
        return metadata

    def domain_split(
        self,
        target_domains: Union[PacsDomain, str, List[Union[str, PacsDomain]]],
    ) -> TrainTestSplit[Self]:
        if isinstance(target_domains, list):
            target_domains = [
                str(str_to_enum(str_=elem, enum=PacsDomain)) for elem in target_domains
            ]
            test_mask = self.metadata["domain"].isin(target_domains)
        else:
            target_domains = str_to_enum(str_=target_domains, enum=PacsDomain)
            test_mask = self.metadata["domain"] == str(target_domains)
        test_mask_np = test_mask.to_numpy()
        test_inds = test_mask_np.nonzero()[0].astype(dtype=np.int64)
        if len(test_inds) == 0:
            raise ValueError(f"No samples in dataset belonging to domain(s) '{target_domains}'.")
        train_inds = (~test_mask_np).nonzero()[0].astype(dtype=np.int64)
        if len(train_inds) == 0:
            raise ValueError(f"No samples in dataset from which to construct the source domain(s).")

        train_subset = self.subset(list(train_inds))
        test_subset = self.subset(list(test_inds))

        return TrainTestSplit(train=train_subset, test=test_subset)



"""CelebA data-module."""

from dataclasses import dataclass
from typing import Any
from typing_extensions import override

import albumentations as A  # type: ignore

from conduit.data.datamodules.vision.base import CdtVisionDataModule
from conduit.data.datasets.vision.celeba import CelebA, CelebASplit, CelebAttr, SampleType
from conduit.data.structures import TrainValTestSplit

__all__ = ["CelebADataModule"]


@dataclass(kw_only=True)
class CelebADataModule(CdtVisionDataModule[SampleType]):
    """Data-module for the CelebA dataset."""

    image_size: int = 224
    superclass: CelebAttr = CelebAttr.SMILING
    subclass: CelebAttr = CelebAttr.MALE
    use_predefined_splits: bool = False

    @override
    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        CelebA(root=self.root, download=True)

    @property
    @override
    def _default_train_transforms(self) -> A.Compose:
        base_transforms = A.Compose(
            [
                A.Resize(self.image_size, self.image_size),
                A.CenterCrop(self.image_size, self.image_size),
            ]
        )
        normalization = super()._default_train_transforms
        return A.Compose([base_transforms, normalization])

    @property
    @override
    def _default_test_transforms(self) -> A.Compose:
        return self._default_train_transforms

    @override
    def _get_image_splits(self) -> TrainValTestSplit[CelebA]:
        # Split the data according to the pre-defined split indices
        if self.use_predefined_splits:
            train_data, val_data, test_data = (
                CelebA(root=self.root, superclass=self.superclass, transform=None, split=split)
                for split in CelebASplit
            )
        # Split the data randomly according to test- and val-prop
        else:
            all_data = CelebA(root=self.root, superclass=self.superclass, transform=None)
            val_data, test_data, train_data = all_data.random_split(
                props=(self.val_prop, self.test_prop),
                seed=self.seed,
            )
        return TrainValTestSplit(train=train_data, val=val_data, test=test_data)



# Generated by configen, do not edit.
# See https://github.com/facebookresearch/hydra/tree/master/tools/configen
# fmt: off
# isort:skip_file
# flake8: noqa

from dataclasses import dataclass, field
from omegaconf import MISSING
from typing import Any


@dataclass
class MemoryBankConf:
    _target_: str = "conduit.models.self_supervised.memory_bank.MemoryBank"
    memory: Any = MISSING  # Tensor



from typing import Callable, List, Optional

import numpy as np
import torch
from torch import Tensor, nn
import torchaudio.transforms as T  # type: ignore

from conduit.data.datasets.utils import AudioTform

__all__ = ["LogMelSpectrogram", "Framing", "Compose"]


class LogMelSpectrogram(T.MelSpectrogram):
    def __init__(
        self,
        sample_rate: int = 16000,
        n_fft: int = 400,
        win_length: Optional[int] = None,
        hop_length: Optional[int] = None,
        f_min: float = 0.0,
        f_max: Optional[float] = None,
        pad: int = 0,
        n_mels: int = 128,
        window_fn: Optional[Callable] = None,
        power: float = 2.0,
        normalized: bool = False,
        wkwargs: Optional[dict] = None,
        center: bool = True,
        pad_mode: str = "reflect",
        onesided: bool = True,
        norm: Optional[str] = None,
        mel_scale: str = "htk",
        log_offset: float = 0.0,
    ):
        if window_fn is None:
            window_fn = torch.hann_window

        super().__init__(
            sample_rate=sample_rate,
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
            f_min=f_min,
            f_max=f_max,
            pad=pad,
            n_mels=n_mels,
            window_fn=window_fn,
            power=power,
            normalized=normalized,
            wkwargs=wkwargs,
            center=center,
            pad_mode=pad_mode,
            onesided=onesided,
            norm=norm,
            mel_scale=mel_scale,
        )
        self.log_offset = log_offset

    def forward(self, waveform: Tensor) -> Tensor:
        x = super().forward(waveform)
        return torch.log(x + self.log_offset)


class Framing(nn.Module):
    def __init__(
        self,
        stft_hop_length_seconds: float = 0.010,
        example_window_seconds: float = 0.96,  # Each example contains 96 10ms frames
        example_hop_seconds: float = 0.96,  # with zero overlap.
    ):
        super().__init__()
        self.stft_hop_length_seconds = stft_hop_length_seconds
        self.example_window_seconds = example_window_seconds
        self.example_hop_seconds = example_hop_seconds

    def forward(self, data: Tensor) -> Tensor:
        # Frame features into examples.
        features_sample_rate = 1.0 / self.stft_hop_length_seconds
        window_length = int(round(self.example_window_seconds * features_sample_rate))
        hop_length = int(round(self.example_hop_seconds * features_sample_rate))

        _, _, num_samples = data.shape
        num_frames = 1 + int(np.floor((num_samples - window_length) / hop_length))
        strides = (1, 1, hop_length)

        return (
            data.unsqueeze(1)
            .as_strided(size=(num_frames, 64, window_length), stride=strides)
            .unsqueeze(1)
            .permute(0, 1, 3, 2)
        )


class Compose(nn.Module):
    def __init__(self, transforms: List[AudioTform]) -> None:
        super().__init__()
        self.transforms = transforms

    def forward(self, audio: Tensor) -> Tensor:
        for tform in self.transforms:
            audio = tform(audio)
        return audio



"""Law Admissions Dataset."""

from dataclasses import dataclass

from ethicml.data import Dataset, Sqf
from ethicml.data import SqfSplits as SqfSens

from conduit.fair.data.datamodules.tabular.base import EthicMlDataModule

__all__ = ["SqfDataModule", "SqfSens"]


@dataclass(kw_only=True)
class SqfDataModule(EthicMlDataModule):
    """NYC Stop, Question, Frisk Dataset."""

    sens_feat: SqfSens = SqfSens.SEX
    disc_feats_only: bool = False

    @property
    def em_dataset(self) -> Dataset:
        return Sqf(split=self.sens_feat, discrete_only=self.disc_feats_only, invert_s=self.invert_s)
